/*
  * Abstraction API/layer for the various ways PvPGN can inspect sockets state
  * 2003 (C) dizzy@roedu.net
  *
  * Code is based on the ideas found in thttpd project.
  *
  * *BSD kqueue() based backend
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include "common/setup_before.h"
#ifdef STDC_HEADERS
# include <stdlib.h>
#else
# ifdef HAVE_MALLOC_H
#  include <malloc.h>
# endif
#endif
#ifdef HAVE_STRING_H
# include <string.h>
#else
# ifdef HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_EVENT_H
# include <sys/event.h>
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#include "fdwatch.h"
#include "common/eventlog.h"
#include "common/setup_after.h"

#ifdef HAVE_KQUEUE
static int sr;
static int kq;
static struct kevent *kqchanges = NULL;		/* changes to make to kqueue */
static struct kevent *kqevents = NULL;		/* events to investigate */
static int *fdw_ridx;
static unsigned nofds;

static int fdw_kqueue_init(int nfds);
static int fdw_kqueue_close(void);
static int fdw_kqueue_add_fd(int fd, t_fdwatch_type rw);
static int fdw_kqueue_del_fd(int fd);
static int fdw_kqueue_watch(long timeout_msecs);
static int fdw_kqueue_check_fd(int fd, t_fdwatch_type rw);
static void fdw_kqueue_handle(void);

t_fdw_backend fdw_kqueue = {
    fdw_kqueue_init,
    fdw_kqueue_close,
    fdw_kqueue_add_fd,
    fdw_kqueue_del_fd,
    fdw_kqueue_watch,
    fdw_kqueue_check_fd,
    fdw_kqueue_handle
};

static int fdw_kqueue_init(int nfds)
{
    int i;

    if ((kq = kqueue()) == -1)
	return -1;
    kqevents = (struct kevent *) malloc(sizeof(struct kevent) * nfds);
    kqchanges = (struct kevent *) malloc(sizeof(struct kevent) * nfds * 2);
    fdw_ridx = (int *) malloc(sizeof(int) * nfds);
    if (fdw_ridx == NULL || kqevents == NULL || kqchanges == NULL)
    {
	fdw_kqueue_close();
	return -1;
    }

    memset(kqchanges, 0, sizeof(struct kevent) * nfds);
    for (i = 0; i < nfds; i++)
	fdw_rw[i] = -1;
    nofds = sr = 0;

    eventlog(eventlog_level_info, __FUNCTION__, "fdwatch kqueue() based layer initialized (max %d sockets)", nfds);
    return 0;
}

static int fdw_kqueue_close(void)
{
    if (fdw_ridx) { free((void *) fdw_ridx); fdw_ridx = NULL; }
    if (kqchanges) { free((void *) kqchanges); kqchanges = NULL; }
    if (kqevents) { free((void *) kqevents); kqevents = NULL; }
    nofds = sr = 0;

    return 0;
}

static int fdw_kqueue_add_fd(int fd, t_fdwatch_type rw)
{
    static int ridx;
    static int filter;

//    eventlog(eventlog_level_trace, __FUNCTION__, "called fd: %d rw: %d", fd, rw);

    filter = 0;
    if (rw & fdwatch_type_read)
	filter |= EVFILT_READ;
    if (rw & fdwatch_type_write)
	filter |= EVFILT_WRITE;

    if (fdw_ridx[fd] < 0 || fdw_ridx[fd] >= nofds || kqchanges[fdw_ridx[fd]].ident != fd)
    {
	ridx = nofds++;
	fdw_ridx[fd] = ridx;
//      eventlog(eventlog_level_trace, __FUNCTION__, "adding new change event fd on %d", ridx);
    } else 
    {
	ridx = fdw_ridx[fd];
//      eventlog(eventlog_level_trace, __FUNCTION__, "updating change event fd on %d", ridx);
    }

    EV_SET(kqchanges + ridx, fd, filter, EV_ADD, 0, 0, 0);

    return 0;
}

static int fdw_kqueue_del_fd(int fd)
{
//    eventlog(eventlog_level_trace, __FUNCTION__, "called fd: %d", fd);
    if (sr > 0) 
	eventlog(eventlog_level_error, __FUNCTION__, "BUG: called while still handling sockets");
    if (nofds && fdw_ridx[fd] < nofds && fdw_ridx[fd] >= 0 && kqchanges[fdw_ridx[fd]].ident == fd)
    {
	/* the last event change about this fd has not yet been sent to kernel */
	nofds--;
	if (fdw_ridx[fd] < nofds)
	{
//          eventlog(eventlog_level_trace, __FUNCTION__, "not last, moving %d", kqchanges[nofds].ident);
	    fdw_ridx[kqchanges[nofds].ident] = fdw_ridx[fd];
	    memcpy(kqchanges + fdw_ridx[fd], kqchanges + nofds, sizeof(struct kevent));
	}
	fdw_ridx[fd] = -1;
    }

/* here we presume the calling code does close() on the socket and if so
 * it is automatically removed from any kernel kqueues */

    return 0;
}

static int fdw_kqueue_watch(long timeout_msec)
{
    static struct timespec ts;

    ts.tv_sec = timeout_msec / 1000L;
    ts.tv_nsec = (timeout_msec % 1000L) * 1000000L;
    sr = kevent(kq, kqchanges, nofds, kqevents, fdw_maxfd, &ts);
    nofds = 0;
    return sr;
}

static int fdw_kqueue_check_fd(int fd, t_fdwatch_type rw)
{
    return 0;
}

static void fdw_kqueue_handle(void)
{
    register unsigned i;

//    eventlog(eventlog_level_trace, __FUNCTION__, "called");
    for (i = 0; i < sr; i++)
    {
//      eventlog(eventlog_level_trace, __FUNCTION__, "checking %d ident: %d read: %d write: %d", i, kqevents[i].ident, kqevents[i].filter & EVFILT_READ, kqevents[i].filter & EVFILT_WRITE);

	if (fdw_rw[kqevents[i].ident] & fdwatch_type_read && kqevents[i].filter & EVFILT_READ)
	    if (fdw_hnd[kqevents[i].ident] (fdw_data[kqevents[i].ident], fdwatch_type_read) == -2)
		continue;

	if (fdw_rw[kqevents[i].ident] & fdwatch_type_write && kqevents[i].filter & EVFILT_WRITE)
	    fdw_hnd[kqevents[i].ident] (fdw_data[kqevents[i].ident], fdwatch_type_write);

    }
    sr = 0;
}

#endif				/* HAVE_KQUEUE */
